@@ -2,7 +2,15 @@ package Catalyst::ActionRole::Methods;
 
 use Moose::Role;
 
-our $VERSION = '0.004';
+our $VERSION = '0.101';
+
+around 'list_extra_info' => sub {
+    my $orig = shift;
+    my $self = shift;
+    my $info = $self->$orig( @_ );
+    $info->{'HTTP_METHOD'} = [ $self->get_allowed_methods( $self->class, undef, $self->name ) ];
+    $info;
+};
 
 around 'dispatch', sub {
     my $orig = shift;
@@ -10,107 +18,63 @@ around 'dispatch', sub {
     my $c    = shift;
 
     my $return = $self->$orig($c, @_);
-    my $rest_method = $self->name . "_" . uc( $c->request->method );
-    my $sub_return = $self->_dispatch_rest_method( $c, $rest_method );
 
-    return defined($sub_return) ? $sub_return : $return;
-};
-
-around 'list_extra_info' => sub {
-  my ($orig, $self, @args) = @_;
-  my @allowed_methods = sort $self->get_allowed_methods($self->class,undef,$self->name);
-  return +{
-    %{ $self->$orig(@args) }, 
-    HTTP_METHODS => \@allowed_methods,
-  };
-};
- 
-sub _dispatch_rest_method {
-    my $self        = shift;
-    my $c           = shift;
-    my $rest_method = shift;
-
-    my $req         = $c->request;
     my $controller = $c->component( $self->class );
-    my ($code, $name);
-  
-    # Common case, for foo_GET etc
-    if ( $code = $controller->action_for($rest_method) ) {
-      return $c->forward( $code,  $req->args ); # Forward to foo_GET if it's an action
+    my $method_name = $self->name;
+    my $req_method = $c->request->method;
+    my $suffix = uc $req_method;
+    my ( $rest_method, $code, $status, $body );
+
+    {
+        $rest_method = $method_name . '_' . $suffix;
+
+        if ( $code = $controller->action_for( $rest_method ) ) {
+            my $sub_return = $c->forward( $code, $c->request->args );
+            return defined $sub_return ? $sub_return : $return;
+        } elsif ( $code = $controller->can( $rest_method ) ) {
+            # nothing to do
+        } elsif ( 'OPTIONS' eq $suffix ) {
+            $status = 204;
+        } elsif ( 'HEAD' eq $suffix ) {
+            $suffix = 'GET';
+            redo;
+        } elsif ( 'not_implemented' eq $suffix ) {
+            $status = 405;
+            $body = '<!DOCTYPE html><title>405 Method Not Allowed</title>'
+                  . "<p>The requested method $req_method is not allowed for this URL.</p>";
+        } else {
+            $suffix = 'not_implemented';
+            redo;
+        }
     }
-    elsif ($code = $controller->can($rest_method)) {
-        $name = $rest_method; # Stash name and code to run 'foo_GET' like an action below.
-    }
- 
-    # Generic handling for foo_*
-    if (!$code) {
-        my $code_action = {
-            OPTIONS => sub {
-                $name = $rest_method;
-                $code = sub { $self->_return_options($self->name, @_) };
-            },
-            HEAD => sub {
-              $rest_method =~ s{_HEAD$}{_GET}i;
-              $self->_dispatch_rest_method($c, $rest_method);
-            },
-            default => sub {
-                # Otherwise, not implemented.
-                $name = $self->name . "_not_implemented";
-                $code = $controller->can($name) # User method
-                    # Generic not implemented
-                    || sub { $self->_return_not_implemented($self->name, @_) };
-            },
-        };
-        my ( $http_method, $action_name ) = ( $rest_method, $self->name );
-        $http_method =~ s{\Q$action_name\E\_}{};
-        my $respond = ($code_action->{$http_method}
-                       || $code_action->{'default'})->();
-        return $respond unless $name;
-    }
- 
+
+    $code ||= defined $status ? sub {
+        my @allowed = $self->get_allowed_methods( $controller, $c, $method_name );
+        $c->response->status( $status );
+        $c->response->header( Allow => @allowed ? \@allowed : '' );
+        if ( defined $body ) {
+            $c->response->content_type( 'text/plain' );
+            $c->response->body( $body );
+        }
+    } : die 'assert: $code || defined $status';
+
     # localise stuff so we can dispatch the action 'as normal, but get
     # different stats shown, and different code run.
     # Also get the full path for the action, and make it look like a forward
     local $self->{code} = $code;
-    my @name = split m{/}, $self->reverse;
-    $name[-1] = $name;
-    local $self->{reverse} = "-> " . join('/', @name);
- 
-    $c->execute( $self->class, $self, @{ $req->args } );
-}
- 
+    ( local $self->{'reverse'} = '-> ' . $self->reverse ) =~ s{[^/]+\z}{$rest_method};
+
+    my $sub_return = $c->execute( $self->class, $self, @{ $c->request->args } );
+    defined $sub_return ? $sub_return : $return;
+};
+
 sub get_allowed_methods {
     my ( $self, $controller, $c, $name ) = @_;
-    my $class = ref($controller) ? ref($controller) : $controller;
-    my $methods = {
-      map { /^$name\_(.+)$/ ? ( $1 => 1 ) : () }
-        ($class->meta->get_all_method_names )
-    };
-    $methods->{'HEAD'} = 1 if $methods->{'GET'};
-    delete $methods->{'not_implemented'};
-    return sort keys %$methods;
-}
- 
-sub _return_options {
-    my ( $self, $method_name, $controller, $c) = @_;
-    my @allowed = $self->get_allowed_methods($controller, $c, $method_name);
-    $c->response->content_type('text/plain');
-    $c->response->status(200);
-    $c->response->header( 'Allow' => @allowed ? \@allowed : '' );
-    $c->response->body(q{});
-}
- 
-sub _return_not_implemented {
-    my ( $self, $method_name, $controller, $c ) = @_;
- 
-    my @allowed = $self->get_allowed_methods($controller, $c, $method_name);
-    $c->response->content_type('text/plain');
-    $c->response->status(405);
-    $c->response->header( 'Allow' => @allowed ? \@allowed : '' );
-    $c->response->body( "Method "
-          . $c->request->method
-          . " not implemented for "
-          . $c->uri_for( $method_name ) );
+    my $class = ref $controller || $controller;
+    my %methods = map /^\Q$name\E\_(.+)()$/, $class->meta->get_all_method_names;
+    $methods{'HEAD'} = 1 if exists $methods{'GET'};
+    delete $methods{'not_implemented'};
+    sort keys %methods;
 }
 
 1;
@@ -172,7 +136,7 @@ class.  The intention is to offer some of the popular functionality that comes
 with L<Catalyst::Action::REST> in a more modular, 'build what you need' package.
 
 Bulk of this documentation and test cases derive from L<Catalyst::Action::REST>
-with the current author's gratitude.
+with the original author's gratitude.
 
 This Action Role handles doing automatic method dispatching for requests.  It
 takes a normal Catalyst action, and changes the dispatch to append an
@@ -197,7 +161,7 @@ For example, in the example above, calling GET on "/foo" would result in
 the foo_GET method being dispatched.
  
 If a method is requested that is not implemented, this action will
-return a status 405 (Method Not Found).  It will populate the "Allow" header
+return a status 405 (Method Not Found).  It will populate the C<Allow> header
 with the list of implemented request methods.  You can override this behavior
 by implementing a custom 405 handler like so:
  
@@ -206,7 +170,7 @@ by implementing a custom 405 handler like so:
    }
  
 If you do not provide an _OPTIONS subroutine, we will automatically respond
-with a 200 OK.  The "Allow" header will be populated with the list of
+with a 204 (No Content). The C<Allow> header will be populated with the list of
 implemented request methods. If you do not provide an _HEAD either, we will
 auto dispatch to the _GET one in case it exists.
 
@@ -303,6 +267,8 @@ mechanism described above.
 
 =head1 AUTHOR
 
+  Aristotle Pagaltzis <pagaltzis@gmx.de>
+
   John Napiorkowski <jnapiork@cpan.org>
 
 Author list from L<Catalyst::Action::REST>
@@ -347,7 +313,7 @@ André Walker (andrewalker) <andre@cpan.org>
  
 =head1 COPYRIGHT
  
-Copyright (c) 2006-2015 the above named AUTHOR and CONTRIBUTORS
+Copyright (c) 2006-2024 the above named authors and contributors
  
 =head1 LICENSE
  
